/* NOTE: this file is autogenerated by version.sh, do not edit */
#ifndef __BUILD_VERSION

#define __BUILD_VERSION

#define BTRFS_LIB_MAJOR 0
#define BTRFS_LIB_MINOR 1
#define BTRFS_LIB_PATCHLEVEL 1

#define BTRFS_LIB_VERSION ( BTRFS_LIB_MAJOR * 10000 + \
                            BTRFS_LIB_MINOR * 100 + \
                            BTRFS_LIB_PATCHLEVEL )

#define BTRFS_BUILD_VERSION "Btrfs v3.17.3"
#endif
